using System;
using System.Collections.Generic;
using MacGruber;
using SimpleJSON;
using UnityEngine;

namespace CheesyFX
{
    public class SlapTrigger : MonoBehaviour
    {
        private TouchZone region;
        private JSONStorableBool enabledJ;
        private JSONStorableAction reset;
        private EventTrigger onUndershot;
        private EventTrigger onExceeded;
        private FloatTrigger onValueChanged;
        private JSONStorableFloat inputTo;
        private JSONStorableFloat inputFrom;
        private JSONStorableFloat threshold;
        public JSONStorableFloat cooldown;
        public JSONStorableFloat decayRate;
        public JSONStorableBool cumulative;
        private UIDynamic decayRateSlider;
        private bool panelOpen;
        private JSONStorableString info = new JSONStorableString("info", "");
        private string baseInfo;
        private float lastValue;

        private float cooldownTimer;
        private float cumulativeValue;

        private bool hasRange = true;
        
        private List<object> UIElements = new List<object>();
        // public Condition condition;

        public void Init(TouchZone region, float thresholdDefault = 5f, float fromDefault = 0f, float toDefault = 100f)
        {
            baseInfo = $"{region.name}: Last Slap Intensity";
            info.val = $"{baseInfo}\n{0:0.00}";
            this.region = region;
            region.slapTrigger = this;
            enabledJ = new JSONStorableBool($"Enabled ({region.name}S)", true);
            reset = new JSONStorableAction($"Reset ({region.name}Slap)", () =>
            {
                cumulativeValue = 0f;
                info.val = $"{baseInfo}\n{0f:0.00}";
            });
            inputFrom = new JSONStorableFloat($"Input From ({region.name}S)", fromDefault, val => hasRange = val != inputTo.val, 0f, 1f, false);
            inputTo = new JSONStorableFloat($"Input To ({region.name}S)", toDefault, val => hasRange = val != inputFrom.val, 0f, 200f, false);
            threshold = new JSONStorableFloat($"Threshold ({region.name}S)", thresholdDefault, 0f, 1f, false);
            cooldown = new JSONStorableFloat($"Cooldown ({region.name}S)", 0f, 0f, 10f, false);
            decayRate = new JSONStorableFloat($"Decay Rate ({region.name}S)", 1f, 0f, 20f, false);
            cumulative = new JSONStorableBool($"Cumulative ({region.name}S)", false);
            onUndershot = new EventTrigger(TouchMe.singleton, "OnExit");
            onExceeded = new EventTrigger(TouchMe.singleton, "OnThresholdExceeded");
            onValueChanged = new FloatTrigger(TouchMe.singleton, "OnValueChanged");

            // condition = new Condition();
            enabledJ.setCallbackFunction += val => enabled = val;
            cooldown.setCallbackFunction += OnResetRateChanged;
            cumulative.setCallbackFunction += OnCumulativeChanged;
            
            TouchMe.singleton.RegisterBool(enabledJ);
            TouchMe.singleton.RegisterAction(reset);
            TouchMe.singleton.RegisterFloat(inputFrom);
            TouchMe.singleton.RegisterFloat(inputTo);
            TouchMe.singleton.RegisterFloat(threshold);
            TouchMe.singleton.RegisterFloat(cooldown);
            TouchMe.singleton.RegisterBool(cumulative);
            TouchMe.singleton.RegisterFloat(decayRate);

            SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
            SuperController.singleton.onAtomAddedHandlers += OnAtomAdded;
        }
        
        private void OnCumulativeChanged(bool val)
        {
            if (!val)
            {
                cumulativeValue = 0f;
                baseInfo = $"{region.name}: Last Slap Intensity";
            }
            else
            {
                cumulativeValue = lastValue;
                baseInfo = $"{region.name}: Cumulative Slap Intensity";
            }
            // Utils.RemoveUIElements(TouchMe.singleton, TriggerUIElements);
            // CreateTriggerUI(TouchMe.singleton);
            decayRate.SetInteractble(val);
            // decayRateSlider.gameObject.SetActive(val);
            decayRateSlider.SetVisible(val);
            info.val = $"{baseInfo}\n{0f:0.00}";
        }

        public void Trigger(float v)
        {
            if (!enabled) return;
            if (cooldownTimer > 0f) return;
            cooldownTimer = cooldown.val;
            if (cumulative.val)
            {
                cumulativeValue += v;
                v = cumulativeValue;
            }

            lastValue = v;
            if (panelOpen) info.val = $"{baseInfo}\n{v:0.00}";
            if(!cumulative.val && onValueChanged.HasActions())
            {
                if (hasRange)
                {
                    onValueChanged.Trigger((v - inputFrom.val) / (inputTo.val - inputFrom.val));

                }
                else onValueChanged.Trigger(v <= inputFrom.val? 0f : 1f);
                // onValueChanged.Update();
            }

            if (v < threshold.val)
            {
                if (onUndershot.HasActions()) onUndershot.Trigger();
            }
            else
            {
                if (onExceeded.HasActions()) onExceeded.Trigger();
            }
        }

        private void Update()
        {
            onUndershot.Update();
            onValueChanged.Update();
            if (cooldownTimer > 0f) cooldownTimer -= Time.deltaTime;
            if (cumulative.val)
            {
                if (cumulativeValue > 0f)
                {
                    if(onValueChanged.HasActions()) onValueChanged.Trigger((cumulativeValue - inputFrom.val) / (inputTo.val - inputFrom.val));
                    cumulativeValue -= decayRate.val * Time.deltaTime;
                }
                if (panelOpen)
                {
                    if (cumulativeValue < 0f) cumulativeValue = 0f;
                    info.val = $"{baseInfo}\n{cumulativeValue:0.00}";
                }
            }
            // if (condition != null && !condition.IsMet()) return;

            // Trigger(region.timeTouched);
            // if (Math.Abs(lastValue - region.timeTouched) > .001f)
            // {
            //     
            //     // region.name.Print();
            // }
        }

        public void OnDestroy()
        {
            onUndershot.Remove();
            onExceeded.Remove();
            onValueChanged.Remove();
            TouchMe.singleton.DeregisterBool(enabledJ);
            TouchMe.singleton.DeregisterAction(reset);
            TouchMe.singleton.DeregisterFloat(inputFrom);
            TouchMe.singleton.DeregisterFloat(inputTo);
            TouchMe.singleton.DeregisterFloat(threshold);
            TouchMe.singleton.DeregisterFloat(cooldown);
            TouchMe.singleton.DeregisterBool(cumulative);
            TouchMe.singleton.DeregisterFloat(decayRate);
            SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;
            SuperController.singleton.onAtomAddedHandlers -= OnAtomAdded;
        }

        public void OnAtomRename(string oldUid, string newUid)
        {
            onUndershot.SyncAtomNames();
            onExceeded.SyncAtomNames();
            onValueChanged.SyncAtomNames();
            // condition.OnAtomRename();
        }

        public void OnAtomAdded(Atom atom)
        {
            // condition.AddExclusionChoice(atom);
        }

        private void OnResetRateChanged(float val)
        {
            if(val > 0f)
            {
                if (!region.touchCollisionListener.isOnStay && region.touchCollisionListener.touchTimerReset == null)
                {
                    region.touchCollisionListener.touchTimerReset =
                        region.touchCollisionListener.TouchTimerReset().Start();
                }
            }
            else if(!region.touchCollisionListener.isOnStay)
            {
                region.touchCollisionListener.touchTimerReset.Stop();
            }
        }

        public JSONClass Store(string subScenePrefix)
        {
            JSONClass jc = new JSONClass();
            inputFrom.Store(jc);
            inputTo.Store(jc);
            threshold.Store(jc);
            cumulative.Store(jc);
            cooldown.Store(jc);
            enabledJ.Store(jc);
            decayRate.Store(jc);
            jc[onExceeded.Name] = onExceeded.GetJSON(subScenePrefix);
            jc[onUndershot.Name] = onUndershot.GetJSON(subScenePrefix);
            jc[onValueChanged.Name] = onValueChanged.GetJSON(subScenePrefix);
            // condition.excludedInfo.Store(jc);
            return jc;
        }

        public void Load(JSONClass jc, string subScenePrefix)
        {
            inputFrom.Load(jc);
            inputTo.Load(jc);
            threshold.Load(jc);
            cumulative.Load(jc);
            cooldown.Load(jc);
            enabledJ.Load(jc);
            decayRate.Load(jc);
            // condition.excludedInfo.Load(jc);
            onExceeded.RestoreFromJSON(jc, subScenePrefix, false, true);
            onUndershot.RestoreFromJSON(jc, subScenePrefix, false, true);
            onValueChanged.RestoreFromJSON(jc, subScenePrefix, false, true);
        }

        public void OpenPanel(MVRScript script, Action back)
        {
            panelOpen = true;
            UIDynamic spacer;
            UIDynamicButton button;
            button = script.CreateButton("Return");
            button.buttonColor = new Color(0.55f, 0.90f, 1f);
            button.button.onClick.AddListener(
                () =>
                {
                    Utils.RemoveUIElements(script, UIElements);
                    panelOpen = false;
                    back();
                });
            UIElements.Add(button);
            
            enabledJ.CreateUI(UIElements: UIElements,rightSide:false);
            cumulative.CreateUI(UIElements: UIElements,rightSide:false);
            cooldown.CreateUI(UIElements: UIElements, rightSide:false);
            
            var textfield = script.CreateTextField(info, true);
            textfield.ForceHeight(115f);
            textfield.UItext.fontSize = 34;
            textfield.UItext.alignment = TextAnchor.MiddleCenter;
            UIElements.Add(textfield);
            
            button = TouchMe.singleton.CreateButton("Reset", true);
            button.button.onClick.AddListener(reset.actionCallback.Invoke);
            UIElements.Add(button);

            // CreateTriggerUI(script);
            decayRateSlider = decayRate.CreateUI(UIElements: UIElements, rightSide:true);
            decayRateSlider.SetVisible(cumulative.val);
            spacer = script.CreateSpacer(true);
            spacer.height = 30f;
            UIElements.Add(spacer);
            
            UIManager.SetupInfoOneLine(UIElements,"Threshold Triggers", false, script: script);

            threshold.CreateUI(script, true, UIElements:UIElements);
            button = script.CreateButton("On Threshold Exceeded", false);
            button.button.onClick.AddListener(onExceeded.OpenPanel);
            button.buttonColor = new Color(0.45f, 1f, 0.45f);
            UIElements.Add(button);
            
            button = script.CreateButton("On Threshold Undershot", false);
            button.button.onClick.AddListener(onUndershot.OpenPanel);
            button.buttonColor = new Color(0.45f, 1f, 0.45f);
            UIElements.Add(button);

            UIManager.SetupInfoOneLine(UIElements,"Value Triggers", false, script: script);
            spacer = script.CreateSpacer(true);
            spacer.height = 30f;
            UIElements.Add(spacer);
            // UIManager.EvenLeftToRight(script, UIElements);
            
            inputFrom.CreateUI(script, UIElements:UIElements);
            inputTo.CreateUI(script, true, UIElements:UIElements);
            
            
            button = script.CreateButton("On Value Changed", false);
            button.button.onClick.AddListener(onValueChanged.OpenPanel);
            button.buttonColor = new Color(0.45f, 1f, 0.45f);
            UIElements.Add(button);



            // if (condition == null) return;
            // UIManager.SetupInfoOneLine(UIElements,"Conditions", true, script: script);
            // // UIManager.EvenLeftToRight();
            // button = script.CreateButton("Exclude...");
            // button.button.onClick.AddListener(() => condition.Exclude(condition.excludedChooser.val, true));
            // UIElements.Add(button);
            // button = script.CreateButton("(Re)Include...");
            // button.button.onClick.AddListener(() => condition.Exclude(condition.excludedChooser.val, false));
            // UIElements.Add(button);
            // condition.excludedChooser.CreateUI(script: script, rightSide: true,chooserType:2, UIElements:UIElements);
            //
            // textfield = script.CreateTextField(condition.excludedInfo, true);
            // textfield.height = 200f;
            // UIElements.Add(textfield);
        }
    }
}