using System;
using System.Collections.Generic;
using System.Linq;
using MacGruber;
using SimpleJSON;
using UnityEngine;

namespace CheesyFX
{
    public class TouchTrigger : MonoBehaviour
    {
        private TouchZone region;
        private JSONStorableBool enabledJ;
        private JSONStorableAction reset;
        private bool absolute;
        private EventTrigger onUndershot;
        private EventTrigger onExceeded;
        private FloatTrigger onValueChanged;
        private JSONStorableFloat inputTo;
        private JSONStorableFloat inputFrom;
        private JSONStorableFloat threshold;
        public JSONStorableFloat decayRate;
        public JSONStorableBool instantReset;
        private bool panelOpen;
        private JSONStorableString info = new JSONStorableString("info", "");
        private string baseInfo;
        private float lastValue;

        private bool hasRange = true;
        
        List<object> UIElements = new List<object>();

        private UIDynamic decayRateSlider;
        // public Condition condition;

        public void Init(TouchZone region, float thresholdDefault = 1f, float fromDefault = 0f, float toDefault = 5f)
        {
            baseInfo = $"{region.name}: Time Touched";
            this.region = region;
            region.touchCollisionListener.touchTrigger = this;
            enabledJ = new JSONStorableBool($"Enabled ({region.name}T)", true);
            reset = new JSONStorableAction($"Reset ({region.name}Touch)", () =>
            {
                region.touchCollisionListener.timeTouched = 0f;
                info.val = $"{baseInfo}\n{0f:0.00}";
            });
            inputFrom = new JSONStorableFloat($"Input From ({region.name}T)", fromDefault, val => hasRange = val != inputTo.val, 0f, 1f, false);
            inputTo = new JSONStorableFloat($"Input To ({region.name}T)", toDefault, val => hasRange = val != inputFrom.val, 0f, 1f, false);
            threshold = new JSONStorableFloat($"Threshold ({region.name}T)", thresholdDefault, 0f, 1f, false);
            decayRate = new JSONStorableFloat($"Decay Rate ({region.name}T)", 10f, 0f, 20f, false);
            instantReset = new JSONStorableBool($"Instant Reset ({region.name}T)", false, val => decayRateSlider.SetVisible(!val));
            onUndershot = new EventTrigger(TouchMe.singleton, "OnUndershot");
            onExceeded = new EventTrigger(TouchMe.singleton, "OnExceeded");
            onValueChanged = new FloatTrigger(TouchMe.singleton, "OnValueChanged");

            // condition = new Condition();

            enabledJ.setCallbackFunction += val => enabled = val;
            decayRate.setCallbackFunction += OnResetRateChanged;
            
            TouchMe.singleton.RegisterBool(enabledJ);
            TouchMe.singleton.RegisterAction(reset);
            TouchMe.singleton.RegisterFloat(inputFrom);
            TouchMe.singleton.RegisterFloat(inputTo);
            TouchMe.singleton.RegisterFloat(threshold);
            TouchMe.singleton.RegisterFloat(decayRate);
            TouchMe.singleton.RegisterBool(instantReset);

            SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
            SuperController.singleton.onAtomAddedHandlers += OnAtomAdded;
        }

        public void Trigger(float v)
        {
            if(onValueChanged.HasActions())
            {
                if (hasRange)
                {
                    onValueChanged.Trigger((v - inputFrom.val) / (inputTo.val - inputFrom.val));

                }
                else onValueChanged.Trigger(v <= inputFrom.val? 0f : 1f);
                // onValueChanged.Update();
            }
            if(onExceeded.HasActions() || onUndershot.HasActions())
            {
                if (lastValue < threshold.val)
                {
                    if (v >= threshold.val) onExceeded.Trigger();
                }
                else if (v < threshold.val) onUndershot.Trigger();
            }
            lastValue = v;
        }

        private void Update()
        {
            onExceeded.Update();
            onUndershot.Update();
            
            // if (condition != null && !condition.IsMet()) return;
            if (panelOpen) info.val = $"{baseInfo}\n{region.timeTouched:0.00}";
            Trigger(region.timeTouched);
            if (Math.Abs(lastValue - region.timeTouched) > .001f)
            {
                onValueChanged.Update();
                // region.name.Print();
            }
        }

        public void OnDestroy()
        {
            onExceeded.Remove();
            onUndershot.Remove();
            onValueChanged.Remove();
            TouchMe.singleton.DeregisterBool(enabledJ);
            TouchMe.singleton.DeregisterAction(reset);
            TouchMe.singleton.DeregisterFloat(inputFrom);
            TouchMe.singleton.DeregisterFloat(inputTo);
            TouchMe.singleton.DeregisterFloat(threshold);
            TouchMe.singleton.DeregisterFloat(decayRate);
            TouchMe.singleton.DeregisterBool(instantReset);
            SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;
            SuperController.singleton.onAtomAddedHandlers -= OnAtomAdded;
        }

        public void OnAtomRename(string oldUid, string newUid)
        {
            onExceeded.SyncAtomNames();
            onUndershot.SyncAtomNames();
            onValueChanged.SyncAtomNames();
            // condition.OnAtomRename();
        }

        public void OnAtomAdded(Atom atom)
        {
            // condition.AddExclusionChoice(atom);
        }

        private void OnResetRateChanged(float val)
        {
            if(val > 0f)
            {
                if (!region.touchCollisionListener.isOnStay && region.touchCollisionListener.touchTimerReset == null)
                {
                    region.touchCollisionListener.touchTimerReset =
                        region.touchCollisionListener.TouchTimerReset().Start();
                }
            }
            else if(!region.touchCollisionListener.isOnStay)
            {
                region.touchCollisionListener.touchTimerReset.Stop();
            }
        }

        public JSONClass Store(string subScenePrefix)
        {
            JSONClass jc = new JSONClass();
            inputFrom.Store(jc);
            inputTo.Store(jc);
            threshold.Store(jc);
            decayRate.Store(jc);
            instantReset.Store(jc);
            enabledJ.Store(jc);
            jc[onExceeded.Name] = onExceeded.GetJSON(subScenePrefix);
            jc[onUndershot.Name] = onUndershot.GetJSON(subScenePrefix);
            jc[onValueChanged.Name] = onValueChanged.GetJSON(subScenePrefix);
            // condition.excludedInfo.Store(jc);
            return jc;
        }

        public void Load(JSONClass jc, string subScenePrefix)
        {
            inputFrom.Load(jc);
            inputTo.Load(jc);
            threshold.Load(jc);
            decayRate.Load(jc);
            instantReset.Load(jc);
            enabledJ.Load(jc);
            // condition.excludedInfo.Load(jc);
            onExceeded.RestoreFromJSON(jc, subScenePrefix, false, true);
            onUndershot.RestoreFromJSON(jc, subScenePrefix, false, true);
            onValueChanged.RestoreFromJSON(jc, subScenePrefix, false, true);
        }

        public void OpenPanel(MVRScript script, Action back)
        {
            panelOpen = true;
            
            UIDynamicButton button;
            button = script.CreateButton("Return");
            button.buttonColor = new Color(0.55f, 0.90f, 1f);
            button.button.onClick.AddListener(
                () =>
                {
                    Utils.RemoveUIElements(script, UIElements);
                    panelOpen = false;
                    back();
                });
            UIElements.Add(button);
            
            enabledJ.CreateUI(UIElements: UIElements,rightSide:false);
            
            decayRateSlider = decayRate.CreateUI(UIElements: UIElements, rightSide:false);
            decayRateSlider.SetVisible(!instantReset.val);

            // Utils.SetupInfoText(script, name, 80f, true);
            var textfield = script.CreateTextField(info, true);
            textfield.ForceHeight(120f);
            textfield.UItext.fontSize = 34;
            textfield.UItext.alignment = TextAnchor.MiddleCenter;
            UIElements.Add(textfield);
            
            instantReset.CreateUI(UIElements: UIElements,rightSide:true);
            
            button = TouchMe.singleton.CreateButton("Reset", true);
            button.button.onClick.AddListener(reset.actionCallback.Invoke);
            UIElements.Add(button);
            
            var spacer = script.CreateSpacer(true);
            spacer.height = 30f;
            UIElements.Add(spacer);
            
            UIManager.SetupInfoOneLine(UIElements,"Threshold Triggers", false, script: script);
            // var spacer = script.CreateSpacer(true);
            // spacer.height = 50f;
            // triggerUIElements.Add(spacer);
            // UIManager.EvenLeftToRight(script, triggerUIElements);

            threshold.CreateUI(script, true, UIElements:UIElements);
            button = script.CreateButton("On Threshold Exceeded", false);
            button.button.onClick.AddListener(onExceeded.OpenPanel);
            button.buttonColor = new Color(0.45f, 1f, 0.45f);
            UIElements.Add(button);
            
            button = script.CreateButton("On Threshold Undershot", false);
            button.button.onClick.AddListener(onUndershot.OpenPanel);
            button.buttonColor = new Color(0.45f, 1f, 0.45f);
            UIElements.Add(button);

            UIManager.SetupInfoOneLine(UIElements,"Value Triggers", false, script: script);
            spacer = script.CreateSpacer(true);
            spacer.height = 30f;
            UIElements.Add(spacer);
            // UIManager.EvenLeftToRight(script, triggerUIElements);
            
            inputFrom.CreateUI(script, UIElements:UIElements);
            inputTo.CreateUI(script, true, UIElements:UIElements);
            
            
            button = script.CreateButton("On Value Changed", false);
            button.button.onClick.AddListener(onValueChanged.OpenPanel);
            button.buttonColor = new Color(0.45f, 1f, 0.45f);
            UIElements.Add(button);
            
            

            // if (condition == null) return;
            // UIManager.SetupInfoOneLine(triggerUIElements,"Conditions", true, script: script);
            // // UIManager.EvenLeftToRight();
            // button = script.CreateButton("Exclude...");
            // button.button.onClick.AddListener(() => condition.Exclude(condition.excludedChooser.val, true));
            // triggerUIElements.Add(button);
            // button = script.CreateButton("(Re)Include...");
            // button.button.onClick.AddListener(() => condition.Exclude(condition.excludedChooser.val, false));
            // triggerUIElements.Add(button);
            // condition.excludedChooser.CreateUI(script: script, rightSide: true,chooserType:2, UIElements:triggerUIElements);
            //
            // textfield = script.CreateTextField(condition.excludedInfo, true);
            // textfield.height = 200f;
            // triggerUIElements.Add(textfield);
        }
    }
}